""" Unit tests for core_linked_records_app.rest.pid_settings.views
"""

from unittest import TestCase
from unittest.mock import patch

from core_linked_records_app.components.pid_settings import (
    api as pid_settings_api,
)
from core_linked_records_app.rest.pid_settings import (
    views as pid_settings_views,
)
from core_linked_records_app.rest.pid_settings.serializers import (
    PidSettingsSerializer,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from tests import mocks
from tests.mocks import MockPidSettings


class TestPidSettingsViewGet(TestCase):
    """Test Pid Settings View Get"""

    def setUp(self):
        """setUp"""

        self.mock_request = mocks.MockRequest()

    @patch.object(pid_settings_api, "get")
    def test_pid_settings_api_get_fails_returns_500(
        self, mock_pid_settings_get
    ):
        """test_pid_settings_api_get_fails_returns_500"""

        mock_pid_settings_get.side_effect = Exception(
            "mock_pid_settings_get_exception"
        )

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.get(self.mock_request)

        self.assertEqual(response.status_code, 500)

    @patch.object(PidSettingsSerializer, "__new__")
    @patch.object(pid_settings_api, "get")
    def test_pid_settings_serializer_fails_returns_500(
        self, mock_pid_settings_get, mock_pid_settings_serializer
    ):
        """test_pid_settings_serializer_fails_returns_500"""

        mock_pid_settings_get.return_value = "mock_pid_settings"
        mock_pid_settings_serializer.side_effect = Exception(
            "mock_pid_settings_serializer_exception"
        )

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.get(self.mock_request)

        self.assertEqual(response.status_code, 500)

    @patch.object(pid_settings_api, "get")
    def test_success_returns_200(self, mock_pid_settings_get):
        """test_success_returns_200"""

        mock_pid_settings_get.return_value = MockPidSettings()

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.get(self.mock_request)

        self.assertEqual(response.status_code, 200)


class TestPidSettingsViewPatch(TestCase):
    """Test Pid Settings View Patch"""

    def setUp(self):
        """setUp"""

        self.mock_request = mocks.MockRequest()
        self.mock_request.user = create_mock_user("1", is_superuser=True)

    @patch.object(PidSettingsSerializer, "__new__")
    def test_pid_serializer_init_fails_returns_500(
        self, mock_pid_settings_serializer
    ):
        """test_pid_serializer_init_fails_returns_500"""

        mock_pid_settings_serializer.side_effect = Exception(
            "mock_pid_settings_serializer_exception"
        )

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.patch(self.mock_request)

        self.assertEqual(response.status_code, 500)

    @patch.object(PidSettingsSerializer, "__new__")
    def test_pid_serializer_is_valid_fails_returns_500(
        self, mock_pid_settings_serializer
    ):
        """test_pid_serializer_is_valid_fails_returns_500"""

        mock_pid_settings_serializer.return_value = mocks.MockSerializer(
            is_valid_exc=Exception("mock_pid_settings_serializer_is_valid_exc")
        )

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.patch(self.mock_request)

        self.assertEqual(response.status_code, 500)

    @patch.object(PidSettingsSerializer, "__new__")
    def test_pid_serializer_not_valid_returns_400(
        self, mock_pid_settings_serializer
    ):
        """test_pid_serializer_not_valid_returns_400"""

        mock_pid_settings_serializer.return_value = mocks.MockSerializer(
            is_valid_result=False
        )

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.patch(self.mock_request)

        self.assertEqual(response.status_code, 400)

    @patch.object(PidSettingsSerializer, "__new__")
    def test_pid_serializer_update_fails_returns_500(
        self, mock_pid_settings_serializer
    ):
        """test_pid_serializer_update_fails_returns_500"""

        mock_pid_settings_serializer.return_value = mocks.MockSerializer(
            is_valid_result=True,
            update_exc=Exception(
                "mock_pid_settings_serializer_update_exception"
            ),
        )

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.patch(self.mock_request)

        self.assertEqual(response.status_code, 500)

    @patch.object(pid_settings_views.PidSettingsView, "get")
    @patch.object(PidSettingsSerializer, "__new__")
    def test_get_fails_returns_500(
        self, mock_pid_settings_serializer, mock_view_get
    ):
        """test_get_fails_returns_500"""

        mock_pid_settings_serializer.return_value = mocks.MockSerializer(
            is_valid_result=True
        )
        mock_view_get.side_effect = Exception("mock_view_get_exception")

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.patch(self.mock_request)

        self.assertEqual(response.status_code, 500)

    @patch.object(pid_settings_views.PidSettingsView, "get")
    @patch.object(PidSettingsSerializer, "__new__")
    def test_success_returns_200(
        self, mock_pid_settings_serializer, mock_view_get
    ):
        """test_success_returns_200"""

        mock_pid_settings_serializer.return_value = mocks.MockSerializer(
            is_valid_result=True
        )
        mock_view_get.return_value = mocks.MockResponse()

        test_view = pid_settings_views.PidSettingsView()
        response = test_view.patch(self.mock_request)

        self.assertEqual(response.status_code, 200)



""" XML utilities functions.
"""

import logging

from xml_utils.commons import exceptions as xml_utils_exceptions
from core_main_app.utils.xml import validate_xml_data
from xml_utils.xpath import create_tree_from_xpath
from xml_utils.xsd_tree.operations.namespaces import (
    get_target_namespace,
    get_namespaces,
)
from xml_utils.xsd_tree.xsd_tree import XSDTree

logger = logging.getLogger(__name__)


def get_xpath_from_dot_notation(dot_notation_path):
    """Transform MongoDB dot notation to XPath

    Params:
        dot_notation_path:

    Returns:

    """
    dot_notation_elements = [
        "{0}:%s" % path if not path.startswith("@") else path
        for path in dot_notation_path.split(".")
    ]

    xpath = "/%s" % "/".join(dot_notation_elements)
    return xpath


def get_target_namespace_for_xsd_string(xsd_string):
    """Retrieve target namespace given a XSD string.

    Args:
        xsd_string:

    Returns:
        dict|None - a dictionary containing namespace name as key, and url as
            value. If no target namespace have been defiend, None is returned.
    """
    xml_tree = XSDTree.build_tree(xsd_string)
    target_namespace, target_namespace_prefix = get_target_namespace(
        xml_tree, get_namespaces(xsd_string)
    )

    if target_namespace_prefix == "":  # No target namespace have been defined
        namespace = None
    else:  # Target namespace is defined
        namespace = {target_namespace_prefix: target_namespace}

    return namespace


def get_xpath_with_target_namespace(xpath, xsd_string):
    """Adds target namespace to a given XPath

    Params:
        xpath:
        xsd_string:

    Returns:

    """
    target_namespace = get_target_namespace_for_xsd_string(xsd_string)

    xpath = xpath.format(
        list(target_namespace.keys())[0]
        if target_namespace is not None
        else ""
    )

    if target_namespace is None:
        xpath = xpath.replace(":", "")

    return xpath


def set_value_at_xpath(xml_tree, xpath, value, namespaces):
    """Set value for a given XPath

    Params:
        xml_tree:
        xpath:
        value:
        namespaces:

    Returns:

    """
    try:
        xml_tree.xpath(xpath, namespaces=namespaces)[0].text = value
    except AttributeError:
        xpath_list = xpath.split("/")
        attribute = xpath_list[-1].replace("@", "")
        xpath = f"/{'/'.join(xpath_list[:-1])}"

        xml_tree.xpath(xpath, namespaces=namespaces)[0].attrib[attribute] = (
            value
        )


def get_value_at_xpath(xml_tree, xpath, namespaces):
    """Retrieve value in XML given a XPath

    Params:
        xml_tree:
        xpath:
        namespaces:

    Returns:

    """
    xpath_element_list = xml_tree.xpath(xpath, namespaces=namespaces)

    # Check that we found exactly one element matching the given xpath
    if len(xpath_element_list) != 1:
        raise xml_utils_exceptions.XPathError(
            f"Expected to find a single element, {len(xpath_element_list)} found"
        )

    try:
        xpath_value = xpath_element_list[0].text
    except AttributeError:  # XPath points to an attribute
        xpath_value = xpath_element_list[0]

    return str(xpath_value) if xpath_value else xpath_value


def can_create_value_at_xpath(xml_string, xsd_string, xpath, value):
    """Evaluate if a value can be set in an XML file at a given XPath

    Params:
        xml_string:
        xsd_string:
        xpath:
        value:

    Returns:
        bool - True if the value can be created, False otherwise.
    """
    try:
        # Replace the current by the modified tree (containing mock PID) and
        # force document PID to be regenerated.
        target_namespace = get_target_namespace_for_xsd_string(xsd_string)
        xml_tree = XSDTree.build_tree(xml_string)

        modified_xml_tree = create_tree_from_xpath(
            xpath, xml_tree, target_namespace
        )
        set_value_at_xpath(modified_xml_tree, xpath, value, target_namespace)
        xsd_tree = XSDTree.build_tree(xsd_string)

        validation_error = validate_xml_data(xsd_tree, modified_xml_tree)
        if validation_error is not None:
            raise Exception(f"Error while validating XML: {validation_error}")

        return True
    except Exception as exc:  # pylint: disable=broad-except
        logger.info("Function 'can_create_value_at_xpath' raised %s", str(exc))
        return False



""" Unit tests for core_linked_records_app.components.oai_record.api
"""

from unittest import TestCase
from unittest.mock import patch

from core_linked_records_app.components.oai_record import (
    api as oai_record_api,
)
from core_linked_records_app.components.pid_path import api as pid_path_api
from core_main_app.commons.exceptions import ApiError
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_oaipmh_harvester_app.components.oai_record import (
    api as oaipmh_harvester_oai_record_api,
)
from tests import mocks


class TestGetPidForData(TestCase):
    """Test Get Pid For Data"""

    def setUp(self) -> None:
        self.mock_request = mocks.MockRequest()
        self.mock_request.user = create_mock_user("1")
        self.kwargs = {
            "oai_record_id": "mock_id",
            "request": self.mock_request,
        }

    @patch.object(oaipmh_harvester_oai_record_api, "get_by_id")
    def test_get_by_id_failure_raises_api_error(self, mock_get_by_id):
        """test_get_by_id_failure_raises_api_error"""

        mock_get_by_id.side_effect = Exception("mock_get_by_id_exception")

        with self.assertRaises(ApiError):
            oai_record_api.get_pid_for_data(**self.kwargs)

    @patch.object(pid_path_api, "get_by_template")
    @patch.object(oaipmh_harvester_oai_record_api, "get_by_id")
    def test_get_by_template_failure_raises_api_error(
        self, mock_get_by_id, mock_get_by_template
    ):
        """test_get_by_template_failure_raises_api_error"""

        mock_get_by_id.return_value = mocks.MockData()
        mock_get_by_template.side_effect = Exception(
            "mock_get_by_template_exception"
        )

        with self.assertRaises(ApiError):
            oai_record_api.get_pid_for_data(**self.kwargs)

    @patch.object(oai_record_api, "get_value_from_dot_notation")
    @patch.object(pid_path_api, "get_by_template")
    @patch.object(oaipmh_harvester_oai_record_api, "get_by_id")
    def test_get_value_from_dot_notation_failure_raises_api_error(
        self,
        mock_get_by_id,
        mock_get_by_template,
        mock_get_value_from_dot_notation,
    ):
        """test_get_value_from_dot_notation_failure_raises_api_error"""

        mock_get_by_id.return_value = mocks.MockData()
        mock_get_by_template.return_value = mocks.MockPidPath()
        mock_get_value_from_dot_notation.side_effect = Exception(
            "mock_get_dict_value_from_key_list_exception"
        )

        with self.assertRaises(ApiError):
            oai_record_api.get_pid_for_data(**self.kwargs)

    @patch.object(oai_record_api, "get_value_from_dot_notation")
    @patch.object(pid_path_api, "get_by_template")
    @patch.object(oaipmh_harvester_oai_record_api, "get_by_id")
    def test_returns_get_value_from_dot_notation(
        self,
        mock_get_by_id,
        mock_get_by_template,
        mock_get_value_from_dot_notation,
    ):
        """test_returns_get_value_from_dot_notation"""

        expected_result = "mock_get_pid_for_data"
        mock_get_by_id.return_value = mocks.MockData()
        mock_get_by_template.return_value = mocks.MockPidPath()
        mock_get_value_from_dot_notation.return_value = expected_result

        self.assertEqual(
            oai_record_api.get_pid_for_data(**self.kwargs),
            expected_result,
        )



""" Unit tests for menus
"""

from unittest import TestCase

from menu import Menu

from core_linked_records_app.menus import pid_settings_menu, admin_menu


class TestMenus(TestCase):
    """Test menus module"""

    def test_pid_settings_menu_in_admin_menu(self):
        """test_pid_settings_menu_in_admin_menu"""
        self.assertIn(pid_settings_menu, admin_menu.children)

    def test_admin_menu_in_loaded_menus(self):
        """test_admin_menu_in_loaded_menus"""
        Menu.load_menus()
        self.assertIn(admin_menu, Menu.items["admin"])



""" Local handle model
"""

from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import RegexValidator
from django.db import models, IntegrityError

from core_main_app.commons import exceptions
from core_main_app.commons.regex import NOT_EMPTY_OR_WHITESPACES


class LocalId(models.Model):
    """Handle object"""

    record_name = models.CharField(
        blank=False,
        unique=True,
        validators=[
            RegexValidator(
                regex=NOT_EMPTY_OR_WHITESPACES,
                message="Title must not be empty or only whitespaces",
                code="invalid_title",
            ),
        ],
        max_length=255,
    )
    record_object_class = models.CharField(blank=True, max_length=255)
    record_object_id = models.CharField(blank=True, max_length=255)

    @staticmethod
    def get_by_name(record_name):
        """Retrieve LocalId object with the given record_name.

        Args:
            record_name:

        Returns:
            LocalId - LocalId object if found.
        """
        try:
            return LocalId.objects.get(  # pylint: disable=no-member
                record_name=record_name
            )
        except ObjectDoesNotExist as dne:
            raise exceptions.DoesNotExist(str(dne))
        except Exception as exc:
            raise exceptions.ModelError(str(exc))

    @staticmethod
    def get_by_class_and_id(record_object_class, record_object_id):
        """Retrieve LocalId object given record_object_class and record_object_id.

        Args:
            record_object_class:
            record_object_id:

        Returns:
            LocalId - LocalId object if found.
        """
        try:
            return LocalId.objects.get(  # pylint: disable=no-member
                record_object_class=record_object_class,
                record_object_id=record_object_id,
            )
        except ObjectDoesNotExist as dne:
            raise exceptions.DoesNotExist(str(dne))
        except Exception as exc:
            raise exceptions.ModelError(str(exc))

    @staticmethod
    def upsert(local_id_object):
        """Insert a new LocalId object

        Args:
            local_id_object:

        Returns:
        """
        try:
            local_id_object.save()
            return local_id_object
        except IntegrityError as nue:
            raise exceptions.NotUniqueError(str(nue))
        except Exception as exc:
            raise exceptions.ModelError(str(exc))

    def __str__(self):
        """LocalId object as string.

        Returns:
            str - String representation of LocalId object.
        """
        return f"LocalId {self.record_name}"



"""Linked records models
"""

from django.db import models

from core_linked_records_app.access_control import rights
from core_main_app.permissions.utils import get_formatted_name


class LinkedRecords(models.Model):
    """LinkedRecords"""

    class Meta:
        """Meta"""

        verbose_name = rights.APP_CONTENT_TYPE
        default_permissions = ()
        permissions = (
            (
                rights.CAN_READ_PID_SETTINGS,
                get_formatted_name(rights.CAN_READ_PID_SETTINGS),
            ),
        )



""" ACL tests for `core_linked_records.components.blob.api`.
"""

from unittest import TestCase
from unittest.mock import patch, MagicMock

from core_linked_records_app.components.blob import (
    api as blob_api,
    access_control as blob_acl,
)
from core_linked_records_app.components.local_id.models import LocalId
from core_linked_records_app.system.local_id import api as local_id_system_api
from core_main_app.access_control import api as main_acl
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from tests import mocks


class TestGetBlobByPid(TestCase):
    """ACL tests for `get_blob_by_pid` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_blob = mocks.MockDocument()

    def setup_mocks(
        self,
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
        user,
        owner=None,
    ) -> None:
        """setup_mocks"""
        mock_valid_pid = "https://websi.te/provider/record"
        mock_get_by_name.return_value = LocalId(
            record_name=mock_valid_pid,
            record_object_class="mock_record_object_class",
            record_object_id="mock_record_object_id",
        )
        mock_import_module.return_value = None

        self.mock_blob.user_id = owner.id if owner else user.id
        mock_blob_module = mocks.MockModule()
        mock_blob_module.get_by_id_return_value = self.mock_blob
        mock_getattr.return_value = mock_blob_module

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    def test_superuser_can_access(
        self,
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_superuser_can_access"""
        user = create_mock_user("1", is_superuser=True)
        self.setup_mocks(
            mock_get_by_name, mock_import_module, mock_getattr, user
        )

        self.assertEqual(
            blob_api.get_blob_by_pid("mock_pid", user), self.mock_blob
        )

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_not_owner_cannot_access_private(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_registered_user_not_owner_cannot_access_private"""

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1")
        owner = create_mock_user("2")
        self.setup_mocks(
            mock_get_by_name, mock_import_module, mock_getattr, user, owner
        )

        with self.assertRaises(AccessControlError):
            blob_api.get_blob_by_pid("mock_pid", user)

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_not_owner_can_access_public(
        self,
        mock_settings,
        mock_workspace_api,
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_registered_user_not_owner_can_access_public"""
        mock_public_workspace = MagicMock()
        self.mock_blob.workspace = mock_public_workspace

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]
        user = create_mock_user("1")
        owner = create_mock_user("2")
        self.setup_mocks(
            mock_get_by_name, mock_import_module, mock_getattr, user, owner
        )

        self.assertEqual(
            blob_api.get_blob_by_pid("mock_pid", user), self.mock_blob
        )

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_and_owner_can_access_private(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_registered_user_and_owner_can_access_private"""

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1")
        self.setup_mocks(
            mock_get_by_name, mock_import_module, mock_getattr, user
        )

        self.assertEqual(
            blob_api.get_blob_by_pid("mock_pid", user), self.mock_blob
        )

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_anonymous_user_not_publc_cannot_access(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_anonymous_user_not_publc_cannot_access"""

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False

        user = create_mock_user("1", is_anonymous=True)

        self.setup_mocks(
            mock_get_by_name, mock_import_module, mock_getattr, user
        )

        with self.assertRaises(AccessControlError):
            blob_api.get_blob_by_pid("mock_pid", user)

    @patch.object(blob_api, "getattr")
    @patch.object(blob_api, "import_module")
    @patch.object(local_id_system_api, "get_by_name")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_anonymous_user_and_public_can_access(
        self,
        mock_settings,
        mock_workspace_api,
        mock_get_by_name,
        mock_import_module,
        mock_getattr,
    ):
        """test_anonymous_user_and_public_can_access"""
        mock_public_workspace = MagicMock()
        self.mock_blob.workspace = mock_public_workspace

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]

        user = create_mock_user("1", is_anonymous=True)
        self.setup_mocks(
            mock_get_by_name, mock_import_module, mock_getattr, user
        )

        self.assertEqual(
            blob_api.get_blob_by_pid("mock_pid", user), self.mock_blob
        )


class TestGetPidForBlob(TestCase):
    """ACL tests for `get_pid_for_blob` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_blob = MagicMock()
        self.mock_blob_pid = "mock_blob_pid"

    def setup_mocks(
        self,
        mock_blob_model,
        mock_blob_system_api,
        user,
        owner=None,
    ) -> None:
        """setup_mocks"""
        self.mock_blob.user_id = owner.id if owner else user.id
        mock_blob_model.get_by_id.return_value = self.mock_blob
        mock_blob_system_api.get_pid_for_blob.return_value = self.mock_blob_pid

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    def test_superuser_can_access(
        self,
        mock_blob_model,
        mock_blob_system_api,
    ):
        """test_superuser_can_access"""
        user = create_mock_user("1", is_superuser=True)

        self.setup_mocks(mock_blob_model, mock_blob_system_api, user)
        self.assertEqual(
            blob_api.get_pid_for_blob("mock_blob_id", user), self.mock_blob_pid
        )

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_not_owner_cannot_access_private(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_blob_model,
        mock_blob_system_api,
    ):
        """test_registered_user_not_owner_cannot_access_private"""
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1", has_perm=False)
        owner = create_mock_user("2")

        self.setup_mocks(mock_blob_model, mock_blob_system_api, user, owner)
        with self.assertRaises(AccessControlError):
            blob_api.get_pid_for_blob("mock_blob_id", user)

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_not_owner_can_access_public(
        self,
        mock_settings,
        mock_workspace_api,
        mock_blob_model,
        mock_blob_system_api,
    ):
        """test_registered_user_not_owner_can_access_public"""
        mock_public_workspace = MagicMock()
        self.mock_blob.workspace = mock_public_workspace

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]
        user = create_mock_user("1")
        owner = create_mock_user("2")

        self.setup_mocks(
            mock_blob_model,
            mock_blob_system_api,
            user,
            owner,
        )
        self.assertEqual(
            blob_api.get_pid_for_blob("mock_blob_id", user), self.mock_blob_pid
        )

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_registered_user_and_owner_can_access_private(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_blob_model,
        mock_blob_system_api,
    ):
        """test_registered_user_and_owner_can_access_private"""
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1")

        self.setup_mocks(mock_blob_model, mock_blob_system_api, user)
        self.assertEqual(
            blob_api.get_pid_for_blob("mock_blob_id", user), self.mock_blob_pid
        )

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_anonymous_user_not_public_cannot_access(
        self,
        mock_settings,
        mock_workspace_api,  # noqa, pylint: disable=unused-argument
        mock_blob_model,
        mock_blob_system_api,
    ):
        """test_anonymous_user_not_public_cannot_access"""
        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        user = create_mock_user("1", is_anonymous=True)

        self.setup_mocks(mock_blob_model, mock_blob_system_api, user)

        with self.assertRaises(AccessControlError):
            blob_api.get_pid_for_blob("mock_blob_id", user)

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    @patch.object(main_acl, "settings")
    def test_anonymous_user_and_public_can_access(
        self,
        mock_settings,
        mock_workspace_api,
        mock_blob_model,
        mock_blob_system_api,
    ):
        """test_anonymous_user_and_public_can_access"""
        mock_public_workspace = MagicMock()
        self.mock_blob.workspace = mock_public_workspace

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = [
            mock_public_workspace
        ]
        user = create_mock_user("1", is_anonymous=True)

        self.setup_mocks(mock_blob_model, mock_blob_system_api, user)
        self.assertEqual(
            blob_api.get_pid_for_blob("mock_blob_id", user), self.mock_blob_pid
        )


class TestSetPidForBlob(TestCase):
    """ACL tests for `set_pid_for_blob` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_blob_pid = "mock_blob_pid"

    def setup_mocks(
        self,
        mock_workspace_api,
        mock_blob,
        mock_blob_system_api,
        user,
        owner=None,
    ) -> None:
        """setup_mocks"""
        mock_workspace = MagicMock()

        mock_workspace_api.get_all_workspaces_with_read_access_by_user.return_value = (
            [mock_workspace] if user == owner else []
        )

        mock_blob_object = MagicMock()
        mock_blob_object.user_id = owner.id if owner else user.id
        mock_blob_object.workspace = mock_workspace

        mock_blob.get_by_id.return_value = mock_blob_object

        mock_blob_system_api.set_pid_for_blob.return_value = self.mock_blob_pid

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    def test_superuser_can_access(
        self,
        mock_workspace_api,
        mock_blob,
        mock_blob_system_api,
    ):
        """test_superuser_can_access"""
        user = create_mock_user("1", is_superuser=True)

        self.setup_mocks(
            mock_workspace_api, mock_blob, mock_blob_system_api, user
        )
        self.assertEqual(
            blob_api.set_pid_for_blob("mock_blob_id", "mock_blob_pid", user),
            self.mock_blob_pid,
        )

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    def test_registered_user_not_owner_cannot_access(
        self,
        mock_workspace_api,
        mock_blob,
        mock_blob_system_api,
    ):
        """test_registered_user_not_owner_cannot_access"""
        user = create_mock_user("1")
        owner = create_mock_user("2")

        self.setup_mocks(
            mock_workspace_api, mock_blob, mock_blob_system_api, user, owner
        )

        with self.assertRaises(AccessControlError):
            blob_api.set_pid_for_blob("mock_blob_id", "mock_blob_pid", user)

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    def test_registered_user_and_owner_can_access(
        self,
        mock_workspace_api,
        mock_blob,
        mock_blob_system_api,
    ):
        """test_registered_user_and_owner_can_access"""
        user = create_mock_user("1")

        self.setup_mocks(
            mock_workspace_api, mock_blob, mock_blob_system_api, user
        )
        self.assertEqual(
            blob_api.set_pid_for_blob("mock_blob_id", "mock_blob_pid", user),
            self.mock_blob_pid,
        )

    @patch.object(blob_api, "blob_system_api")
    @patch.object(blob_acl, "Blob")
    @patch.object(main_acl, "workspace_api")
    def test_anonymous_user_cannot_access(
        self,
        mock_workspace_api,
        mock_blob,
        mock_blob_system_api,
    ):
        """test_anonymous_user_without_perm_cannot_access"""
        user = create_mock_user("1", is_anonymous=True)

        self.setup_mocks(
            mock_workspace_api, mock_blob, mock_blob_system_api, user
        )

        with self.assertRaises(AccessControlError):
            blob_api.set_pid_for_blob("mock_blob_id", "mock_blob_pid", user)



""" Initialize permissions related to PID implementation.
"""

import logging

from core_linked_records_app.access_control import rights
from core_main_app.components.group import api as group_api
from core_main_app.permissions import (
    rights as main_rights,
    api as permissions_api,
)

logger = logging.getLogger(__name__)


def init_permissions():
    """Initialize groups and permissions."""
    try:
        # Get or create the default group
        default_group, _ = group_api.get_or_create(
            name=main_rights.DEFAULT_GROUP
        )

        can_read_pid_settings_perm = permissions_api.get_by_codename(
            codename=rights.CAN_READ_PID_SETTINGS
        )

        # Add permissions to default group.
        default_group.permissions.add(
            can_read_pid_settings_perm,
        )
    except Exception as exception:  # pylint: disable=broad-except
        logger.error(
            "Impossible to init permissions for core_linked_records_app: %s",
            str(exception),
        )



""" Unit tests for data_html_renderer packages.
"""

from unittest import TestCase

from rest_framework import status

from core_linked_records_app.rest.pid.views import DataHtmlRenderByPID
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock


class TestDataHtmlRenderByPID(TestCase):
    """TestDataHtmlRenderByPID"""

    def test_get_data_html_render_by_pid_returns_http_400_response_when_pid_is_missing(
        self,
    ):
        """test_get_data_html_render_by_pid_returns_http_400_response_when_pid_is_missing"""

        mock_user = create_mock_user("1")

        response = RequestMock.do_request_get(
            DataHtmlRenderByPID.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)



""" Permission tests for core_linked_records_app.rest.pid_settings.views
"""

from unittest import TestCase
from unittest.mock import patch, Mock

from rest_framework import status

from core_linked_records_app.components.pid_settings import (
    api as pid_settings_api,
)
from core_linked_records_app.rest.pid_settings import (
    views as pid_settings_views,
)
from core_linked_records_app.rest.pid_settings.serializers import (
    PidSettingsSerializer,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock
from tests import mocks


class TestPidSettingsViewGet(TestCase):
    """Test Pid Settings View Get"""

    @patch.object(PidSettingsSerializer, "__new__")
    @patch.object(pid_settings_api, "get")
    def test_anonymous_returns_403(
        self, mock_pid_settings_get, mock_pid_setting_settings_serializer
    ):
        """test_anonymous_returns_403"""

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_pid_setting_settings_serializer.return_value = Mock(data={})

        response = RequestMock.do_request_get(
            pid_settings_views.PidSettingsView.as_view(), None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(PidSettingsSerializer, "__new__")
    @patch.object(pid_settings_api, "get")
    def test_authenticated_returns_200(
        self, mock_pid_settings_get, mock_pid_setting_settings_serializer
    ):
        """test_authenticated_returns_200"""

        mock_user = create_mock_user("1")

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_pid_setting_settings_serializer.return_value = Mock(data={})

        response = RequestMock.do_request_get(
            pid_settings_views.PidSettingsView.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(PidSettingsSerializer, "__new__")
    @patch.object(pid_settings_api, "get")
    def test_staff_returns_200(
        self, mock_pid_settings_get, mock_pid_setting_settings_serializer
    ):
        """test_staff_returns_200"""

        mock_user = create_mock_user("1", is_staff=True)

        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        mock_pid_setting_settings_serializer.return_value = Mock(data={})

        response = RequestMock.do_request_get(
            pid_settings_views.PidSettingsView.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestPidSettingsViewPatch(TestCase):
    """Test Pid Settings View Patch"""

    def test_anonymous_returns_403(self):
        """test_anonymous_returns_403"""

        response = RequestMock.do_request_patch(
            pid_settings_views.PidSettingsView.as_view(), None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_403(self):
        """test_authenticated_returns_403"""

        mock_user = create_mock_user("1")

        response = RequestMock.do_request_patch(
            pid_settings_views.PidSettingsView.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_staff_returns_403(self):
        """test_staff_returns_403"""

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_patch(
            pid_settings_views.PidSettingsView.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(pid_settings_api, "get")
    @patch.object(PidSettingsSerializer, "__new__")
    def test_superuser_returns_200(
        self, mock_pid_serializer, mock_pid_settings_get
    ):
        """test_staff_returns_200"""

        mock_user = create_mock_user("1", is_superuser=True)

        mock_pid_serializer.return_value = mocks.MockSerializer()
        mock_pid_settings_get.return_value = mocks.MockPidSettings()
        response = RequestMock.do_request_patch(
            pid_settings_views.PidSettingsView.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)



""" Handle system abstract class
"""

import logging
from abc import ABC, abstractmethod
from importlib import import_module

from django.urls import reverse
from rest_framework import status

from core_linked_records_app import settings
from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import CoreError

logger = logging.getLogger(__name__)


class AbstractIdProvider(ABC):
    """Abstract Id Provider"""

    def __init__(self, provider_name, provider_lookup_url):
        core_linked_records_provider_records = reverse(
            "core_linked_records_provider_record",
            kwargs={"provider": provider_name, "record": ""},
        )
        self.local_url = (
            f"{settings.SERVER_URI}" f"{core_linked_records_provider_records}"
        )

        self.local_url = self.local_url[:-1]  # Removing the final /
        self.provider_lookup_url = (
            provider_lookup_url if provider_lookup_url else self.local_url
        )

    @abstractmethod
    def get(self, record):
        """get
        Args:
            record:
        """
        raise NotImplementedError()

    @abstractmethod
    def create(self, prefix, record=None):
        """create
        Args:
            prefix:
            record:
        """
        raise NotImplementedError()

    @abstractmethod
    def update(self, record):
        """update
        Args:
            record:
        """
        raise NotImplementedError()

    @abstractmethod
    def delete(self, record):
        """delete
        Args:
            record:
        """
        raise NotImplementedError()


class ProviderManager:
    """Manage provider instances from a given provider name"""

    def __init__(self):
        self._provider_instance = None

        self.provider_name = settings.ID_PROVIDER_SYSTEM_NAME
        self.provider_config = settings.ID_PROVIDER_SYSTEM_CONFIG

    def get(self, provider_name=None) -> AbstractIdProvider:
        """get provider
        Args:
            provider_name:

        Returns:

        """
        if self._provider_instance is None:
            # Retrieve class name and module path
            id_provider_classpath = self.provider_config["class"].split(".")
            id_provider_classname = id_provider_classpath[-1]
            id_provider_modpath = ".".join(id_provider_classpath[:-1])

            # Import module and class
            id_provider_module = import_module(id_provider_modpath)
            id_provider_class = getattr(
                id_provider_module, id_provider_classname
            )

            # Initialize handle system instance, default to the system specified in
            # settings.
            if provider_name is None:
                provider_name = settings.ID_PROVIDER_SYSTEM_NAME

            self._provider_instance = id_provider_class(
                provider_name, *settings.ID_PROVIDER_SYSTEM_CONFIG["args"]
            )

        return self._provider_instance

    def find_provider_from_pid(self, pid):
        """find_provider_from_pid
        Args:
            pid:

        Returns:

        """
        provider = self.get(settings.ID_PROVIDER_SYSTEM_NAME)

        return (
            settings.ID_PROVIDER_SYSTEM_NAME
            if pid.startswith(provider.provider_lookup_url)
            else None
        )


def retrieve_provider_name(pid_value):
    """Retrieve name of the provider given a PID.

    Args:
        pid_value:

    Returns:
         str - Provider name
    """

    if pid_value is None or pid_value == "":  # PID field left blank
        # Select the default provider if no PID has been chosen.
        provider_name = settings.ID_PROVIDER_SYSTEM_NAME
    else:  # PID specified in document.
        provider_name = ProviderManager().find_provider_from_pid(pid_value)

    # Detect provider specified by PID. Raise an error if no matching providers are
    # found.
    if provider_name is None:
        raise exceptions.ModelError(
            "Invalid PID provided (provider not found)"
        )

    return provider_name


def delete_record_from_provider(record_name: str):
    """Delete a PID from the provider using the record name as stored in the
    LocalId table.

    Args:
        record_name: str - Formatted as prefix/record.

    Raises:
        ApiError: if the HTTP status is not 200 or 404.
    """
    # Delete the given LocalID fron the default PID provider.
    provider = ProviderManager().get()
    previous_pid_delete_response = provider.delete(record_name)

    if previous_pid_delete_response.status_code == status.HTTP_200_OK:
        return

    # At this point, there was some problem deleting the LocalID needed to be logged.
    error_message = (
        f"Deletion of LocalID {record_name} from provider "
        f"{ProviderManager().provider_name} returned "
        f"{previous_pid_delete_response.status_code}"
    )

    # Do not crash for a 404, but log that an error occured.
    if previous_pid_delete_response.status_code == status.HTTP_404_NOT_FOUND:
        logger.warning(error_message)
        return

    # Log any error that happened during PID deletion.
    logger.error(error_message)
    raise CoreError(error_message)



""" ACL tests for `core_linked_records.components.pid_path.api`.
"""

from unittest import TestCase
from unittest.mock import patch, MagicMock

from core_linked_records_app.components.pid_path import (
    api as pid_path_api,
    models as pid_path_models,
)
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.components.template import (
    access_control as template_access_control,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user


class TestGetByTemplate(TestCase):
    """ACL tests for `get_by_template` function."""

    def setUp(self) -> None:
        """setUp"""
        self.mock_template = MagicMock()
        self.mock_pid_path = MagicMock()

    def setup_mocks(self, owner) -> None:
        """setup_mocks"""
        self.mock_template.user = owner.id if owner else None

    @patch.object(pid_path_models.PidPath, "get_by_template")
    def test_superuser_can_access(self, mock_get_by_template):
        """test_superuser_can_access"""
        user = create_mock_user("1", is_superuser=True)

        mock_get_by_template.return_value = self.mock_pid_path
        self.setup_mocks(user)

        self.assertEqual(
            pid_path_api.get_by_template(self.mock_template, user),
            self.mock_pid_path,
        )

    @patch.object(pid_path_models.PidPath, "get_by_template")
    def test_registered_user_not_owner_cannot_access_private(
        self, mock_get_by_template
    ):
        """test_registered_user_without_perm_cannot_access"""
        user = create_mock_user("1")
        owner = create_mock_user("2")

        mock_get_by_template.return_value = self.mock_pid_path
        self.setup_mocks(owner)

        with self.assertRaises(AccessControlError):
            pid_path_api.get_by_template(self.mock_template, user)

    @patch.object(pid_path_models.PidPath, "get_by_template")
    def test_registered_user_not_owner_can_access_public(
        self, mock_get_by_template
    ):
        """test_registered_user_without_perm_cannot_access"""
        user = create_mock_user("1")

        mock_get_by_template.return_value = self.mock_pid_path
        self.setup_mocks(None)  # noqa

        self.assertEqual(
            pid_path_api.get_by_template(self.mock_template, user),
            self.mock_pid_path,
        )

    @patch.object(pid_path_models.PidPath, "get_by_template")
    def test_registered_user_and_owner_can_access_private(
        self, mock_get_by_template
    ):
        """test_registered_user_with_perm_can_access"""
        user = create_mock_user("1")

        mock_get_by_template.return_value = self.mock_pid_path
        self.setup_mocks(user)

        self.assertEqual(
            pid_path_api.get_by_template(self.mock_template, user),
            self.mock_pid_path,
        )

    @patch.object(pid_path_models.PidPath, "get_by_template")
    @patch.object(template_access_control, "settings")
    def test_anonymous_user_cannot_access_private(
        self, mock_settings, mock_get_by_template
    ):
        """test_anonymous_user_without_perm_cannot_access"""
        user = create_mock_user("1", is_anonymous=True)

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = False
        mock_get_by_template.return_value = self.mock_pid_path
        self.setup_mocks(None)  # noqa

        with self.assertRaises(AccessControlError):
            pid_path_api.get_by_template(self.mock_template, user)

    @patch.object(pid_path_models.PidPath, "get_by_template")
    @patch.object(template_access_control, "settings")
    def test_anonymous_user_can_access_public(
        self, mock_settings, mock_get_by_template
    ):
        """test_anonymous_user_with_perm_can_access"""
        user = create_mock_user("1", is_anonymous=True)

        mock_settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = True
        mock_get_by_template.return_value = self.mock_pid_path
        self.setup_mocks(None)  # noqa

        self.assertEqual(
            pid_path_api.get_by_template(self.mock_template, user),
            self.mock_pid_path,
        )



""" Unit tests for `core_linked_records.components.data.access_control`.
"""

from unittest import TestCase
from unittest.mock import MagicMock, patch

from core_linked_records_app.components.data import access_control as data_acl
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.utils.tests_tools.MockUser import create_mock_user


class TestCanGetDataByPid(TestCase):
    """Unit tests for `can_get_data_by_pid` function."""

    def setUp(self) -> None:
        """setUp"""
        mock_request = MagicMock()
        mock_request.user = create_mock_user("1")

        self.mock_kwargs = {
            "func": MagicMock(),
            "pid": "mock_pid",
            "request": mock_request,
        }

    @patch.object(data_acl, "check_can_read_document")
    def test_func_with_args_called(
        self,
        mock_check_can_read_document,  # noqa, pylint: disable=unused-argument
    ):
        """test_func_with_args_called"""
        data_acl.can_get_data_by_pid(**self.mock_kwargs)

        self.mock_kwargs["func"].assert_called_with(
            self.mock_kwargs["pid"],
            self.mock_kwargs["request"],
        )

    @patch.object(data_acl, "check_can_read_document")
    def test_check_can_read_document_not_called_for_superuser(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_not_called_for_superuser"""
        self.mock_kwargs["request"].user = create_mock_user(
            "1", is_superuser=True
        )
        data_acl.can_get_data_by_pid(**self.mock_kwargs)

        mock_check_can_read_document.assert_not_called()

    @patch.object(data_acl, "check_can_read_document")
    def test_check_can_read_document_called(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_called"""
        data_acl.can_get_data_by_pid(**self.mock_kwargs)

        mock_check_can_read_document.assert_called_with(
            self.mock_kwargs["func"](
                self.mock_kwargs["pid"],
                self.mock_kwargs["request"],
            ),
            self.mock_kwargs["request"].user,
        )

    @patch.object(data_acl, "check_can_read_document")
    def test_check_can_read_document_acl_error_fails(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_called"""
        mock_check_can_read_document.side_effect = AccessControlError(
            "mock_check_can_read_document_acl_error"
        )

        with self.assertRaises(AccessControlError):
            data_acl.can_get_data_by_pid(**self.mock_kwargs)

    @patch.object(data_acl, "check_can_read_document")
    def test_succesful_operation_returns_data(
        self,
        mock_check_can_read_document,  # noqa, pylint: disable=unused-argument
    ):
        """test_succesful_operation_returns_data"""
        self.assertEqual(
            data_acl.can_get_data_by_pid(**self.mock_kwargs),
            self.mock_kwargs["func"](
                self.mock_kwargs["pid"],
                self.mock_kwargs["request"],
            ),
        )


class TestCanGetPidForData(TestCase):
    """Unit tests for `can_get_pid_for_data` function."""

    def setUp(self) -> None:
        """setUp"""
        mock_request = MagicMock()
        mock_request.user = create_mock_user("1")

        self.mock_kwargs = {
            "func": MagicMock(),
            "data_id": "mock_data_id",
            "request": mock_request,
        }

    @patch.object(data_acl, "check_can_read_document")
    @patch.object(data_acl, "Data")
    def test_check_can_read_document_called(
        self,
        mock_data,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_called"""
        mock_data_object = MagicMock()

        mock_data.get_by_id.return_value = mock_data_object
        data_acl.can_get_pid_for_data(**self.mock_kwargs)

        mock_check_can_read_document.assert_called_with(
            mock_data_object, self.mock_kwargs["request"].user
        )

    @patch.object(data_acl, "check_can_read_document")
    def test_check_can_read_document_not_called_for_superuser(
        self,
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_not_called_for_superuser"""
        self.mock_kwargs["request"].user = create_mock_user(
            "1", is_superuser=True
        )
        data_acl.can_get_pid_for_data(**self.mock_kwargs)

        mock_check_can_read_document.assert_not_called()

    @patch.object(data_acl, "check_can_read_document")
    @patch.object(data_acl, "Data")
    def test_check_can_read_document_acl_error_fails(
        self,
        mock_data,  # noqa, pylint: disable=unused-argument
        mock_check_can_read_document,
    ):
        """test_check_can_read_document_called"""
        mock_check_can_read_document.side_effect = AccessControlError(
            "mock_check_can_read_document_acl_error"
        )

        with self.assertRaises(AccessControlError):
            data_acl.can_get_pid_for_data(**self.mock_kwargs)

    @patch.object(data_acl, "check_can_read_document")
    @patch.object(data_acl, "Data")
    def test_succesful_operation_returns_func_with_args(
        self,
        mock_data,  # noqa, pylint: disable=unused-argument
        mock_check_can_read_document,  # noqa, pylint: disable=unused-argument
    ):
        """test_succesful_operation_returns_func_with_args"""
        self.assertEqual(
            data_acl.can_get_pid_for_data(**self.mock_kwargs),
            self.mock_kwargs["func"](
                self.mock_kwargs["data_id"],
                self.mock_kwargs["request"],
            ),
        )



""" Local resolver Blob API.
"""

from importlib import import_module
from logging import getLogger

from core_linked_records_app.components.blob.access_control import (
    can_get_blob_by_pid,
    can_get_pid_for_blob,
    can_set_pid_for_blob,
)
from core_linked_records_app.system.blob import api as blob_system_api
from core_linked_records_app.system.local_id import api as local_id_system_api
from core_main_app.access_control.decorators import access_control
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons import exceptions

logger = getLogger(__name__)


@access_control(can_get_blob_by_pid)
def get_blob_by_pid(pid, user):
    """Return blob object with the given pid.

    Args:
        pid (str): PID of the blob.
        user (User): User making the request, should have access to the Blob.

    Raises:
        DoesNotExist: The PID is not assigned to any Blob object.
        AccessControlError: The Blob cannot be accessed by the user.
        ApiError: Any other error occured while trying to resolve the PID.

    Returns:
        Blob: Blob object assigned to the given PID.
    """
    try:
        # From the PID url (e.g. https://pid-system.org/prefix/record), retrieve
        # only the prefix and record (e.g. prefix/record) stored in DB.
        pid_internal_name = "/".join(pid.split("/")[-2:])
        local_id_object = local_id_system_api.get_by_name(pid_internal_name)

        # Ensure the LocalID object refers to a Blob
        assert (
            local_id_object.record_object_class
            and local_id_object.record_object_id
        )

        # From the local ID object, retrieve record module, separating import path
        # and module name.
        record_object_classpath = local_id_object.record_object_class
        record_object_module_path = ".".join(
            record_object_classpath.split(".")[:-1]
        )
        api_module_name = record_object_classpath.split(".")[-1]

        # Import record module and call 'get_by_id' function with record ID as
        # parameter.
        module = import_module(record_object_module_path)
        return getattr(module, api_module_name).get_by_id(
            local_id_object.record_object_id, user
        )
    except (AssertionError, exceptions.DoesNotExist) as not_assigned_exc:
        error_message = f"PID '{pid}' not assigned to blob"
        logger.error(error_message)
        raise exceptions.DoesNotExist(error_message) from not_assigned_exc
    except AccessControlError as acl_error:
        logger.error(str(acl_error))
        raise AccessControlError(str(acl_error)) from acl_error
    except Exception as exc:
        error_message = (
            f"An error occurred while looking up blob assigned to PID '{pid}'"
        )

        logger.error("%s: %s", error_message, str(exc))
        raise exceptions.ApiError(error_message) from exc


@access_control(can_get_pid_for_blob)
def get_pid_for_blob(blob_id, user):  # noqa, pylint: disable=unused-argument
    """Retrieve PID matching the blob ID provided.

    Args:
        blob_id (str): Primary key of the Blob object.
        user (User): User making the request, should have access to the Blob object.

    Returns:
        str - PID of the blob object
    """
    return blob_system_api.get_pid_for_blob(blob_id)


@access_control(can_set_pid_for_blob)
def set_pid_for_blob(
    blob_id, blob_pid, user  # noqa, pylint: disable=unused-argument
):
    """Create a PID for the blob ID provided.

    Args:
        blob_id (str): Primary key of the Blob object.
        blob_pid (str): PID to assign to the Blob.
        user (User): User making the request, should have proper permissions to access
            the Blob and assign a PID

    Returns:
        LocalId: PID object created for the given Blob object.
    """
    return blob_system_api.set_pid_for_blob(blob_id, blob_pid)



""" Integration tests for data watchers (creation / deletion of PIDs when data are
created / deleted).
"""

from os.path import join

from core_linked_records_app.settings import (
    ID_PROVIDER_PREFIX_DEFAULT,
    ID_PROVIDER_SYSTEM_NAME,
)
from core_linked_records_app.utils import (
    exceptions as linked_records_exceptions,
)
from core_main_app.commons import exceptions as main_exceptions
from core_main_app.utils.integration_tests.integration_base_transaction_test_case import (
    IntegrationTransactionTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from tests.fixtures import DataFixtures
from tests.test_settings import SERVER_URI


class TestRecordCreationWithDuplicatePid(IntegrationTransactionTestCase):
    """Integration tests checking the creation of two data with duplicate PID."""

    fixture = DataFixtures()

    def setUp(self):  # pylint: disable=invalid-name
        """setUp"""
        self.user = create_mock_user(1)
        self.mock_pid = "pid1"
        self.mock_pid_url = join(
            SERVER_URI,
            "rest",
            ID_PROVIDER_SYSTEM_NAME,
            ID_PROVIDER_PREFIX_DEFAULT,
            self.mock_pid,
        )
        super().setUp()

    def test_create_duplicate_pid_is_not_possible_when_pid_enabled(
        self,
    ):
        """test_create_duplicate_pid_is_not_possible_when_pid_enabled"""
        # Arrange: Enable PID.
        self.fixture.auto_set_pid(True)

        # Act: Load data 1 with PID 1.
        self.fixture.insert_record("record_1", self.mock_pid_url, self.user)

        # Assert: Ensure creating data 2 with PID 1 raises an exception.
        with self.assertRaises(
            main_exceptions.ModelError,
            msg="PID already defined for another instance",
        ):
            self.fixture.insert_record(
                "record_2", self.mock_pid_url, self.user
            )

    def test_create_duplicate_pid_is_possible_when_pid_disabled(self):
        """test_create_duplicate_pid_is_possible_when_pid_disabled"""
        # Arrange: Disable PID
        self.fixture.auto_set_pid(False)

        # Act / Assert: Load data 1 with PID 1 then data 2 with PID 1. No exception
        #   should occur.
        self.fixture.insert_record("record_1", self.mock_pid_url, self.user)
        self.fixture.insert_record("record_2", self.mock_pid_url, self.user)

    def test_create_duplicate_pid_is_not_possible_after_pid_enabled(
        self,
    ):
        """test_create_duplicate_pid_is_not_possible_after_pid_enabled"""
        # Arrange: Disable PID.
        self.fixture.auto_set_pid(False)

        # Act: Load data 1 with PID 1 then enable PIDs.
        self.fixture.insert_record("record_1", self.mock_pid_url, self.user)
        self.fixture.auto_set_pid(True)

        # Assert: Ensure creating data 2 with PID 1 raises an exception.
        with self.assertRaises(
            main_exceptions.ModelError,
            msg="PID already defined for another instance",
        ):
            self.fixture.insert_record(
                "record_2", self.mock_pid_url, self.user
            )

    def test_create_duplicate_pid_is_possible_after_pid_disabled(
        self,
    ):
        """test_create_duplicate_pid_is_possible_after_pid_disabled"""
        # Arrange: Enable PID.
        self.fixture.auto_set_pid(True)

        # Act / Assert: Load data 1 with PID 1, disable PIDs and, create data 2 with
        #   PID 1. No exception should occur.
        self.fixture.insert_record("record_1", self.mock_pid_url, self.user)
        self.fixture.auto_set_pid(False)
        self.fixture.insert_record("record_2", self.mock_pid_url, self.user)


class TestRecordCreationWithDeletedPid(IntegrationTransactionTestCase):
    """Integration tests checking the creation of a data containing a deleted
    PID."""

    fixture = DataFixtures()

    def setUp(self):  # pylint: disable=invalid-name
        """setUp"""
        self.user = create_mock_user(1)
        self.mock_pid = "pid1"
        self.mock_pid_url = join(
            SERVER_URI,
            "rest",
            ID_PROVIDER_SYSTEM_NAME,
            ID_PROVIDER_PREFIX_DEFAULT,
            self.mock_pid,
        )
        super().setUp()

    def test_create_record_is_possible_when_pid_enabled(self):
        """test_create_record_is_possible_when_pid_enabled"""
        self.fixture.auto_set_pid(True)

        data_1 = self.fixture.insert_record(
            "record_1", self.mock_pid_url, self.user
        )
        data_1.delete()

        # Act / Assert: Creating record should not raise any exception.
        self.fixture.insert_record("record_2", self.mock_pid_url, self.user)

    def test_create_record_pid_is_possible_when_pid_disabled(self):
        """test_create_record_pid_is_possible_when_pid_disabled"""
        # Arrange: Disable PID + create & delete record.
        self.fixture.auto_set_pid(False)

        data_1 = self.fixture.insert_record(
            "record_1", self.mock_pid_url, self.user
        )
        data_1.delete()

        # Act / Assert: Creating record should not raise any exception.
        self.fixture.insert_record("record_2", self.mock_pid_url, self.user)

    def test_create_record_pid_is_possible_after_pid_enabled(
        self,
    ):
        """test_create_record_pid_is_possible_after_pid_enabled"""
        # Arrange: Disable PID + create & delete data.
        self.fixture.auto_set_pid(False)

        data_1 = self.fixture.insert_record(
            "record_1", self.mock_pid_url, self.user
        )
        data_1.delete()

        # Act / Assert: Enable PID + check creating a new record with the same PID does
        #   not raise any exception.
        self.fixture.auto_set_pid(True)
        self.fixture.insert_record("record_2", self.mock_pid_url, self.user)

    def test_create_record_pid_is_possible_after_pid_disabled(
        self,
    ):
        """test_create_record_pid_is_possible_after_pid_disabled"""
        # Arrange: Enable PID + create & delete record.
        self.fixture.auto_set_pid(True)

        data_1 = self.fixture.insert_record(
            "record_1", self.mock_pid_url, self.user
        )
        data_1.delete()

        # Act / Assert: Disable PID + check creating a new record with the same PID does
        #   not raise any exception.
        self.fixture.auto_set_pid(False)
        self.fixture.insert_record("record_2", self.mock_pid_url, self.user)


class TestRecordModificationToNewPid(IntegrationTransactionTestCase):
    """Integration tests checking the modification of a data, changing the existing PID
    to a new, unregistered, PID.
    """

    fixture = DataFixtures()

    def setUp(self):  # pylint: disable=invalid-name
        """setUp"""
        self.user = create_mock_user(1)
        self.mock_pid_1 = "pid1"
        self.mock_pid_url_1 = join(
            SERVER_URI,
            "rest",
            ID_PROVIDER_SYSTEM_NAME,
            ID_PROVIDER_PREFIX_DEFAULT,
            self.mock_pid_1,
        )
        self.mock_pid_2 = "pid2"
        self.mock_pid_url_2 = join(
            SERVER_URI,
            "rest",
            ID_PROVIDER_SYSTEM_NAME,
            ID_PROVIDER_PREFIX_DEFAULT,
            self.mock_pid_2,
        )
        super().setUp()

    def test_modify_record_is_possible_when_pid_enabled(
        self,
    ):
        """test_modify_record_is_possible_when_pid_enabled"""
        # Arrange: Enable PID + create record_1 with PID 1.
        self.fixture.auto_set_pid(True)
        data_1 = self.fixture.insert_record(
            "record_1", self.mock_pid_url_1, self.user
        )

        # Act / Assert: Modify record_1 to PID 2 should not raise any exception.
        data_1.xml_content = data_1.xml_content.replace(
            self.mock_pid_url_1, self.mock_pid_url_2
        )
        data_1.save()

    def test_modify_record_pid_is_possible_when_pid_disabled(self):
        """test_modify_record_pid_is_possible_when_pid_disabled"""
        # Arrange: Disable PID + create record_1 with PID 1.
        self.fixture.auto_set_pid(False)
        data_1 = self.fixture.insert_record(
            "record_1", self.mock_pid_url_1, self.user
        )

        # Act / Assert: Modify record_1 to PID 2 should not raise any exception.
        data_1.xml_content = data_1.xml_content.replace(
            self.mock_pid_url_1, self.mock_pid_url_2
        )
        data_1.save()

    def test_modify_record_pid_is_possible_after_pid_enabled(
        self,
    ):
        """test_modify_record_pid_is_possible_after_pid_enabled"""
        # Arrange: Disable PID + create record_1 with PID 1.
        self.fixture.auto_set_pid(False)
        data_1 = self.fixture.insert_record(
            "record_1", self.mock_pid_url_1, self.user
        )

        # Act / Assert: Enable PID & check that modifying record_1 to PID 2 does not
        #   raise any exception.
        self.fixture.auto_set_pid(True)
        data_1.xml_content = data_1.xml_content.replace(
            self.mock_pid_url_1, self.mock_pid_url_2
        )
        data_1.save()

    def test_modify_record_pid_is_possible_after_pid_disabled(
        self,
    ):
        """test_modify_record_pid_is_possible_after_pid_disabled"""
        # Arrange: Enable PID + create record_1 with PID 1.
        self.fixture.auto_set_pid(True)
        data_1 = self.fixture.insert_record(
            "record_1", self.mock_pid_url_1, self.user
        )

        # Act / Assert: Disable PID & check that modifying record_1 to PID 2 does not
        #   raise any exception.
        self.fixture.auto_set_pid(False)
        data_1.xml_content = data_1.xml_content.replace(
            self.mock_pid_url_1, self.mock_pid_url_2
        )
        data_1.save()


class TestRecordModificationToExistingPid(IntegrationTransactionTestCase):
    """Integration tests checking the modification of a data, changing the existing PID
    to an already registered PID.
    """

    fixture = DataFixtures()

    def setUp(self):  # pylint: disable=invalid-name
        """setUp"""
        self.user = create_mock_user(1)
        self.mock_pid_1 = "pid1"
        self.mock_pid_url_1 = join(
            SERVER_URI,
            "rest",
            ID_PROVIDER_SYSTEM_NAME,
            ID_PROVIDER_PREFIX_DEFAULT,
            self.mock_pid_1,
        )
        self.mock_pid_2 = "pid2"
        self.mock_pid_url_2 = join(
            SERVER_URI,
            "rest",
            ID_PROVIDER_SYSTEM_NAME,
            ID_PROVIDER_PREFIX_DEFAULT,
            self.mock_pid_2,
        )
        super().setUp()

    def test_modify_record_is_not_possible_when_pid_enabled(
        self,
    ):
        """test_modify_record_is_not_possible_when_pid_enabled"""
        # Arrange: Enable PID + create record_1 with PID 1 and record_2 with PID 2.
        self.fixture.auto_set_pid(True)
        self.fixture.insert_record("record_1", self.mock_pid_url_1, self.user)
        data_2 = self.fixture.insert_record(
            "record_2", self.mock_pid_url_2, self.user
        )

        # Act: Change data_2 pid to PID 1.
        data_2.xml_content = data_2.xml_content.replace(
            self.mock_pid_url_2, self.mock_pid_url_1
        )

        # Assert: Saving the data should raise an error.
        with self.assertRaises(
            linked_records_exceptions.PidCreateError,
            msg="PID already defined for another instance",
        ):
            data_2.save()

    def test_modify_record_pid_is_possible_when_pid_disabled(self):
        """test_modify_record_pid_is_possible_when_pid_disabled"""
        # Arrange: Disalbe PID + create record_1 with PID 1 and record_2 with PID 2
        self.fixture.auto_set_pid(False)
        self.fixture.insert_record("record_1", self.mock_pid_url_1, self.user)
        data_2 = self.fixture.insert_record(
            "record_2", self.mock_pid_url_2, self.user
        )

        # Act: Change data_2 pid to PID 1.
        data_2.xml_content = data_2.xml_content.replace(
            self.mock_pid_url_2, self.mock_pid_url_1
        )

        # Assert: Saving the data should not raise an error.
        data_2.save()

    def test_modify_record_pid_is_not_possible_after_pid_enabled(
        self,
    ):
        """test_modify_record_pid_is_not_possible_after_pid_enabled"""
        # Arrange: Disable PID + create record_1 with PID 1 and record_2 with PID 2
        self.fixture.auto_set_pid(False)
        self.fixture.insert_record("record_1", self.mock_pid_url_1, self.user)
        data_2 = self.fixture.insert_record(
            "record_2", self.mock_pid_url_2, self.user
        )

        # Act: Enable PID + Change data_2 pid to PID 1
        self.fixture.auto_set_pid(True)
        data_2.xml_content = data_2.xml_content.replace(
            self.mock_pid_url_2, self.mock_pid_url_1
        )

        # Assert: Saving the data should raise an error.
        with self.assertRaises(
            linked_records_exceptions.PidCreateError,
            msg="PID already defined for another instance",
        ):
            data_2.save()

    def test_modify_record_pid_is_possible_after_pid_disabled(
        self,
    ):
        """test_modify_record_pid_is_possible_after_pid_disabled"""
        # Arrange: Enable PID + create record_1 with PID 1 and record_2 with PID 2
        self.fixture.auto_set_pid(True)
        self.fixture.insert_record("record_1", self.mock_pid_url_1, self.user)
        data_2 = self.fixture.insert_record(
            "record_2", self.mock_pid_url_2, self.user
        )

        # Act: Disable PID + Change data_2 pid to PID 1
        self.fixture.auto_set_pid(False)
        data_2.xml_content = data_2.xml_content.replace(
            self.mock_pid_url_2, self.mock_pid_url_1
        )

        # Assert: Saving the data should not raise an error.
        data_2.save()
